/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query 
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse)
{
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse))
    {
        // 处理表名
        query->tables = std::move(x->tabs);
        /** TODO: 检查表是否存在 */

        for (const auto &tab : query->tables) {
            if (!(sm_manager_->db_.is_table(tab))) {
                throw TableNotFoundError(tab);
            }
        }

        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);

        //处理列名
        get_sel_col(x->cols,query->cols,query->agre_cols,all_cols);

//        for (auto &order_col: query->order_cols){
//            order_col = check_column(all_cols, order_col);
//        }
        //此时已经处理好字段名
        //处理where条件
        get_clause(x->conds, query->conds,query->has_sublink);
        check_clause(query->tables, query->conds);//TODO 校验having是否存在

        //获取groupby
        get_groupby_col(x->group_by_cols,query->groupby_cols);
        check_groupby_clause(query->groupby_cols,query->cols);
        //处理having条件
        get_having_clause(x->having_conds,query->having_conds);
        check_having_clause(query->tables,query->having_conds);
        // infer table name from column name for having_condition_agrecol
        for (auto &sel_col : query->having_conds) {
            if(sel_col.agre_lhs_col.agre_col.col_name.empty() &&sel_col.agre_lhs_col.agre_col.tab_name.empty())continue;//处理COUNT(*)，保证表名和列名为空
            sel_col.agre_lhs_col.agre_col = check_column(all_cols, sel_col.agre_lhs_col.agre_col);  // 列元数据校验
        }

    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        /** TODO: */

        // 检查表格是否存在
        auto tab = x->tab_name;
        if (!(sm_manager_->db_.is_table(tab))) {
            throw TableNotFoundError(tab);
        }
        query->tables.push_back(tab);

        // 取出set_clause并放入query相应的数组中
        for (const auto &clause : x->set_clauses) {
            auto col_name = clause->col_name;

            TabCol tab_col = {tab, col_name};
            Value val = convert_sv_value(clause->val);
            UpdateOp op = convert_sv_update_op(clause->op);
            query->set_clauses.push_back({tab_col, val, op});
        }

        //处理where条件
        get_clause(x->conds, query->conds,query->has_sublink);
        check_clause({x->tab_name}, query->conds);
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        //处理where条件
        get_clause(x->conds, query->conds,query->has_sublink);
        check_clause({x->tab_name}, query->conds);        
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        // 处理insert 的values值
        for (auto &sv_val : x->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}

/**
 *
 * @param sv_cols 语法树解析得到的目标列，select 后的所有部分，包括聚合列
 * @param regular_col 表中原有列
 * @param agre_col 聚合产生的新列
 * @brief 目的，得到正确的列名，分离聚合列和正常列
 */
void Analyze::get_sel_col(const std::vector<std::shared_ptr<ast::Col>> &sv_cols, std::vector<TabCol> &regular_col,
                          std::vector<AgreCol> &agre_col,const std::vector<ColMeta> &all_cols) {
    for(auto &sv_sel_col : sv_cols){
        if(!sv_sel_col->is_aggregate){
            TabCol sel_col = {.tab_name = sv_sel_col->tab_name, .col_name = sv_sel_col->col_name};
            regular_col.push_back(sel_col);
            continue;
        }
        //deal with count(*)
        AgreType tmp_agre_type;
        TabCol tmp_alias;
        TabCol tmp_agre_col;
        if(sv_sel_col->col_name.empty() && sv_sel_col->agre_type == ast::SvAgreType::TYPE_COUNT){
            tmp_agre_type = AgreType::AGRE_TYPE_COUNT_ALL;
        }else{
            tmp_agre_type = convert_sv_agretype(sv_sel_col->agre_type);
        }
        tmp_alias =  {.tab_name = sv_sel_col->tab_name,.col_name = sv_sel_col->alias};
        tmp_agre_col = {.tab_name = sv_sel_col->tab_name, .col_name = sv_sel_col->col_name};//应该还有一个推断过程，判断是哪一个表
        AgreCol agre_info = {.alias = tmp_alias,.agre_col = tmp_agre_col,.agre_type = tmp_agre_type};
        agre_col.push_back(agre_info);
    }
    if (regular_col.empty() && agre_col.empty()) {
        // select all columns
        for (auto &col : all_cols) {
            TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name};
            regular_col.push_back(sel_col);
        }
    } else {
        // infer table name from column name for regular select
        for (auto &sel_col : regular_col) {
            sel_col = check_column(all_cols, sel_col);  // 列元数据校验
        }
        //infer table name from column name for agre select
        for(auto &sel_col : agre_col){
            if(sel_col.agre_col.col_name.empty() &&sel_col.agre_col.tab_name.empty())continue;//处理COUNT(*)，保证表名和列名为空
            sel_col.agre_col = check_column(all_cols, sel_col.agre_col);  // 列元数据校验
            sel_col.alias.tab_name = sel_col.agre_col.tab_name;
        }

    }
}

void Analyze::get_groupby_col(const std::vector<std::shared_ptr<ast::Col>> &sv_groupyby_cols,
                              std::vector<TabCol> &groupby_cols) {
    for(auto &sv_groupby_col : sv_groupyby_cols){
        TabCol tmp = {.tab_name = sv_groupby_col->tab_name,.col_name = sv_groupby_col->col_name};
        groupby_cols.push_back(tmp);
    }
}

void Analyze::get_having_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_having_conds,
                                std::vector<AgreCondition> &having_conds) {
    having_conds.clear();
    for(auto &sv_having_cond : sv_having_conds){
        AgreCondition having_cond;
        having_cond.op = convert_sv_comp_op(sv_having_cond->op);
        TabCol agre_col = {.tab_name = sv_having_cond->lhs->tab_name,.col_name = sv_having_cond->lhs->col_name};
        //deal with count(*) agre_type
        AgreType tmp_agre_type;
        if(agre_col.col_name.empty() && sv_having_cond->lhs->agre_type == ast::SvAgreType::TYPE_COUNT){
            tmp_agre_type = AgreType::AGRE_TYPE_COUNT_ALL;
        }else{
            tmp_agre_type = convert_sv_agretype(sv_having_cond->lhs->agre_type);
        }
        having_cond.agre_lhs_col.agre_col = agre_col;
        having_cond.agre_lhs_col.agre_type = tmp_agre_type;
//        having_cond.agre_lhs_col = {.agre_col.col_name = agre_col.col_name,
//                                    .agre_col.tab_name = agre_col.tab_name,
//                                    .agre_type = tmp_agre_type};//////

        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(sv_having_cond->rhs)) {
            having_cond.rhs_val = convert_sv_value(rhs_val);
        }else{
            assert(0);//右值，只允许数字
            //TODO 除了数字，还可以是其他聚合值
        }
        having_conds.push_back(having_cond);
    }
}

TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target) {

    if(!target.tab_name.empty()){
        /** TODO: Make sure target column exists */
        auto tab_name = target.tab_name;
        auto col_name = target.col_name;

        bool found = false;
        for (const auto &col : all_cols) {
            if (col.tab_name == tab_name && col.name == col_name) {
                if (!found) {
                    found = true;
                } else {
                    throw AmbiguousColumnError(col_name);
                }
            }
        }

        if (!found) {
            throw ColumnNotFoundError(col_name);
        }
        return target;
    }

    // Table name not specified, infer table name from column name

    std::string tab_name;
    for (auto &col : all_cols) {
        if (col.name == target.col_name) {
            if (!tab_name.empty()) {
                throw AmbiguousColumnError(target.col_name);
            }
            tab_name = col.tab_name;
        }
    }
    if (tab_name.empty()) {
        throw ColumnNotFoundError(target.col_name);
    }
    target.tab_name = tab_name;

    return target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names, std::vector<ColMeta> &all_cols) {
    for (auto &sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}

void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<Condition> &conds,bool &has_sublink) {
    conds.clear();
    has_sublink = false;
    for (auto &expr : sv_conds) {
        if(expr->lhs->is_aggregate)throw IncompatibleCondition();
        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.op = convert_sv_comp_op(expr->op);
        cond.is_rhs_sub_query = false;
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }else if(auto rhs_subquery = std::dynamic_pointer_cast<ast::SubQueryExpr>(expr->rhs)){
            cond.is_rhs_val = true;
            if(rhs_subquery->is_const){
                for(auto value : rhs_subquery->const_values){
                    auto tmp = convert_sv_value(value);
                    cond.rhs_val.add_to_set(tmp);
                }
            }else{
                cond.is_rhs_sub_query = true;
                std::shared_ptr<Query> subquery = do_analyze(rhs_subquery->subquery); // 解析子查询
                // Store the subquery or its results in the condition
                cond.sub_query_info = std::make_shared<SubQueryInfo>();
                cond.sub_query_info->sub_query = subquery;
                has_sublink = true;
            }
        }
        conds.push_back(cond);
    }
}

void Analyze::check_clause(const std::vector<std::string> &tab_names, std::vector<Condition> &conds) {
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &cond : conds) {
        // Infer table name from column name
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if(cond.is_rhs_sub_query){
            if(cond.op != OP_IN){
                //标量子查询
                if(cond.sub_query_info->sub_query->agre_cols.size() != 1){
                    throw IncompatibleOpError();
                }
            }else{
                //集合查询
                //throw IncompatibleOpError();
            }
            return;
        }

        if (!cond.is_rhs_val) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }
        TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        if (cond.is_rhs_val && !cond.is_rhs_sub_query) {
            cond.rhs_val.init_raw(lhs_col->len);
            rhs_type = cond.rhs_val.type;
        } else if(cond.is_rhs_val && cond.is_rhs_sub_query) {
            auto subquery_selcols = cond.sub_query_info->sub_query->cols;
            if(subquery_selcols.size() > 1)throw InvalidValueCountError();
            auto subquery_selcol = subquery_selcols[0];//只返回单列，多列需要报错
            TabMeta &rhs_tab = sm_manager_->db_.get_table(subquery_selcol.tab_name);
            auto rhs_col = rhs_tab.get_col(subquery_selcol.col_name);
            rhs_type = rhs_col->type;
        }else{
            TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }
        if (lhs_type != rhs_type) {
            if (!((lhs_type==TYPE_INT || lhs_type==TYPE_FLOAT)
                 &&(rhs_type==TYPE_FLOAT || rhs_type==TYPE_INT))) {
                throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
            }
        }
    }
}

void Analyze::check_groupby_clause(const std::vector<TabCol> &groupby_cols,std::vector<TabCol> regular_cols) {

    if(groupby_cols.empty()){
        return;
    }
    //SELECT 列表中不能出现没有在 GROUP BY 子句中的非聚集列
    for(auto &regular_col : regular_cols){
        bool find = false;
        for(auto groupby_col : groupby_cols){
            if(regular_col.col_name == groupby_col.col_name){
                find = true;
            }
        }
        if(!find){
            throw ColumnNotFoundError(regular_col.col_name);
        }
    }

}

void Analyze::check_having_clause(const std::vector<std::string> &tab_names, std::vector<AgreCondition> &having_conds) {
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // infer table name from column name for having_condition_agrecol
    for (auto &sel_col : having_conds) {
        if(sel_col.agre_lhs_col.agre_col.col_name.empty() &&sel_col.agre_lhs_col.agre_col.tab_name.empty())continue;//处理COUNT(*)，保证表名和列名为空
        sel_col.agre_lhs_col.agre_col = check_column(all_cols, sel_col.agre_lhs_col.agre_col);  // 列元数据校验

    }

}


Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
        {ast::SV_OP_IN, OP_IN}
    };
    return m.at(op);
}
AgreType Analyze::convert_sv_agretype(ast::SvAgreType type){
    std::cout << "Converting SvAgreType: " << static_cast<int>(type) << std::endl;
    std::map<ast::SvAgreType, AgreType> m = {
            {ast::SvAgreType::TYPE_MAX, AGRE_TYPE_MAX}, {ast::SvAgreType::TYPE_MIN, AGRE_TYPE_MIN},
            {ast::SvAgreType::TYPE_SUM, AGRE_TYPE_SUM}, {ast::SvAgreType::TYPE_COUNT, AGRE_TYPE_COUNT},
            {ast::SvAgreType::TYPE_COUNT_ALL,AgreType::AGRE_TYPE_COUNT_ALL}
    };
    return m.at(type);
}

UpdateOp Analyze::convert_sv_update_op(ast::SvUpdateOp op) {
    if (op == ast::SvUpdateOp::SV_ADD) {
        return UpdateOp::PLUS;
    } else if (op == ast::SvUpdateOp::SV_MINUS) {
        return UpdateOp::MINUS;
    } else if (op == ast::SvUpdateOp::SV_MULTI) {
        return UpdateOp::MULTI;
    } else if (op == ast::SvUpdateOp::SV_DIVIDE) {
        return UpdateOp::DIVIDE;
    }
    return UpdateOp::NO;
}
