/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SortExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;
    ColMeta cols_;                              // 框架中只支持一个键排序，需要自行修改数据结构支持多个键排序
    size_t tuple_num;
    bool is_desc_;
    std::vector<size_t> used_tuple;
    std::unique_ptr<RmRecord> current_tuple;

    std::vector<std::unique_ptr<RmRecord>> all_tuples;
    int size_;
    int current_index_;

   public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, TabCol sel_cols, bool is_desc) {
        prev_ = std::move(prev);
        cols_ = prev_->get_col_offset(sel_cols);
        is_desc_ = is_desc;
        tuple_num = 0;
        used_tuple.clear();
        size_ = 0;
        current_index_ = 0;
    }

    void beginTuple() override { 
        prev_->beginTuple();
        while (!prev_->is_end()) {
            auto record = prev_->Next();
            all_tuples.push_back(std::move(record));
            size_++;
            prev_->nextTuple();
        }

        std::sort(all_tuples.begin(), all_tuples.end(), [this](std::unique_ptr<RmRecord>& record1, std::unique_ptr<RmRecord>& record2) {
            auto left_value = get_value(record1, cols_);
            auto right_value = get_value(record2, cols_);
            if (is_desc_) {
                return compare_value(left_value, right_value, OP_GE);
            }
            return compare_value(left_value, right_value, OP_LE);
        });
        current_index_ = 0;
    }

    void nextTuple() override {
        if (is_end()) {
            return;
        }
        current_index_++;
    }

    std::unique_ptr<RmRecord> Next() override {
        if (is_end()) {
            return nullptr;
        }
        auto record = std::move(all_tuples[current_index_]);
        return record;
    }

    bool check_condition(std::unique_ptr<RmRecord> &record1, std::unique_ptr<RmRecord> &record2) {
        auto left_value = get_value(record1, cols_);
        auto right_value = get_value(record2, cols_);

        if (is_desc_) {
            return compare_value(left_value, right_value, OP_GE);
        }
        return compare_value(left_value, right_value, OP_LE);
    }

    bool compare_value(const Value& left_value,const Value& right_value,CompOp op){
        switch (op){
            case OP_EQ:
                return left_value == right_value;
            case OP_NE:
                return left_value != right_value;
            case OP_LT:
                return left_value < right_value;
            case OP_GT:
                return left_value > right_value;
            case OP_LE:
                return !(left_value > right_value);
            case OP_GE:
                return !(left_value < right_value);
        }
    }

    Rid &rid() override { return _abstract_rid; }

    bool is_end() const override { return current_index_ == size_; }

    const std::vector<ColMeta> &cols() const override {
        return prev_->cols();
    }
};