/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;                      // 表名称
    TabMeta tab_;                               // 表的元数据
    std::vector<Condition> conds_;              // 扫描条件
    RmFileHandle *fh_;                          // 表的数据文件句柄
    std::vector<ColMeta> cols_;                 // 需要读取的字段
    size_t len_;                                // 选取出来的一条记录的长度
    std::vector<Condition> fed_conds_;          // 扫描条件，和conds_字段相同

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    std::unique_ptr<RecScan> scan_;

    SmManager *sm_manager_;

    bool isend;

    Iid left_;
    Iid right_;
    Iid gap_left_;
    Iid gap_right_;
    std::string index_name_;

public:
    IndexScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, std::vector<std::string> index_col_names,
                    Context *context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        conds_ = std::move(conds);
        // index_no_ = index_no;
        index_col_names_ = index_col_names; 
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        for (auto &cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                // lhs is on other table, now rhs must be on this table
                assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
                // swap lhs and rhs
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
        }
        fed_conds_ = conds_;
        isend = false;
        index_name_ = sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta_.cols);
        context_->lock_mgr_->lock_IS_on_table(context_->txn_, fh_->GetFd());
    }

    //获得过滤后的索引的边界
    //condition中可能包含某一索引列，根据那一列进行过滤，得到b+树的最左和最右的slot情况
    //对于多索引，以第一个非等号符号为OP，对于一个索引的多个<, <=,获取其相关最小值的符号,若值相等，则取<
    void getBound() {
        char *lower = new char[index_meta_.col_tot_len];
        char *upper = new char[index_meta_.col_tot_len];
        bool set_lower = false, set_upper = false;
        // 遍历索引中所有的列，将lower和upper填满
        int current_index = 0;
        CompOp min_cmp = OP_NONE;
        CompOp max_cmp = OP_NONE;//用于记录id<=2还是id<2的符号大小情况，在多维索引中，以第一个列为基准。
        for (const auto &index_col : index_meta_.cols) {
            auto index_col_name = index_col.name;
            CompOp index_min_cmp = CompOp::OP_NONE;
            CompOp index_max_cmp = CompOp::OP_NONE;
            Value current_min, current_max; // 初始化最大值和最小值
            // 初始化
            init_current_min_max(current_min,current_max,index_col);

            for (const auto &cond : conds_) {
                if(cond.lhs_col.col_name != index_col_name || cond.lhs_col.tab_name != tab_name_)continue;
                //获得索引的值的边界的大小
                switch(cond.op){
                    case OP_EQ:
                        if (current_min < cond.rhs_val) {
                            current_min = cond.rhs_val;
                            set_lower = true;

                        }
                        if (current_max > cond.rhs_val) {
                            current_max = cond.rhs_val;
                            set_upper = true;
                        }
                        index_min_cmp = (index_min_cmp == OP_NONE)?OP_EQ:index_min_cmp;
                        index_max_cmp = (index_max_cmp == OP_NONE)?OP_EQ:index_max_cmp;
                        break;
                    case OP_GE:
                        if (current_min < cond.rhs_val) {
                            current_min = cond.rhs_val;
                            set_lower = true;
                            index_min_cmp = OP_GE;
                        }
                        break;
                    case OP_GT:
                        if (current_min < cond.rhs_val || current_min == cond.rhs_val) {
                            current_min = cond.rhs_val;
                            index_min_cmp = OP_GT;
                            set_lower = true;
                        }
                        break;
                    case OP_LE:
                        if (current_max > cond.rhs_val) {
                            current_max = cond.rhs_val;
                            set_upper = true;
                            index_max_cmp = OP_LE;
                        }
                        break;
                    case OP_LT:
                        if (current_max > cond.rhs_val || current_max == cond.rhs_val) {
                            current_max = cond.rhs_val;
                            set_upper = true;
                            index_max_cmp = OP_LT;
                        }
                        break;
                    default:
                        assert(0);
                }

            }
            min_cmp = (min_cmp == CompOp::OP_EQ || min_cmp == CompOp::OP_NONE)?index_min_cmp:min_cmp;
            max_cmp = (max_cmp == CompOp::OP_EQ || max_cmp == CompOp::OP_NONE)?index_max_cmp:max_cmp;

            memcpy(lower + current_index, current_min.raw->data, index_col.len);
            memcpy(upper + current_index, current_max.raw->data, index_col.len);
            current_index += index_col.len;
        }

        auto index_file_handle = sm_manager_->ihs_.at(index_name_).get();
        //等值查询不用间隙锁
        if(min_cmp == OP_EQ && max_cmp == OP_EQ){
            left_ = index_file_handle->lower_bound(lower);
            Rid rid;
            right_ = index_file_handle->upper_bound(upper,rid);
            return;
        }
        //left gap获取
        if(min_cmp == OP_NONE || min_cmp == OP_EQ){
            left_ = index_file_handle->leaf_begin();
            gap_left_ = index_file_handle->leaf_begin();
        }else{
            switch (min_cmp) {
                case OP_GE:{
                    left_ = index_file_handle->lower_bound(lower);
                    Rid rid;
                    gap_left_ = index_file_handle->upper_bound(lower,rid);
                    break;
                }
                case OP_GT:{
                    left_ = index_file_handle->lower_bound(lower);
                    Rid rid;
                    gap_left_ = index_file_handle->upper_bound(lower,rid);
                    break;
                }
                default:
                    assert(0);
            }
        }

        //right gap获取
        if(max_cmp == OP_NONE || max_cmp == OP_EQ){
            right_ = index_file_handle->leaf_end();
            gap_right_ = index_file_handle->leaf_end();
        }else{
            switch (max_cmp) {
                case OP_LT:{
                    //Rid rid;
                    right_ =  index_file_handle->upper_bound(upper);
                    gap_right_ = index_file_handle->lower_bound(upper);
                    break;
                }
                case OP_LE:{
                    Rid rid;
                    right_ = index_file_handle->upper_bound(upper);
                    gap_right_ = index_file_handle->lower_bound(upper);
                    break;
                }
                default:
                    assert(0);
            }
        }

        //update_left_right_by_cmp(cmp,set_lower,set_upper,lower,upper);
//        left_ = set_lower ? sm_manager_->ihs_.at(index_name_)->lower_bound(lower) : sm_manager_->ihs_.at(index_name_)->leaf_begin();
//        right_ = set_upper ? sm_manager_->ihs_.at(index_name_)->upper_bound(upper) : sm_manager_->ihs_.at(index_name_)->leaf_end();
//        gap_left_ = set_lower ? sm_manager_->ihs_.at(index_name_)->upper_bound(lower) : sm_manager_->ihs_.at(index_name_)->leaf_begin();
//        gap_right_ = set_upper ? sm_manager_->ihs_.at(index_name_)->lower_bound(upper) : sm_manager_->ihs_.at(index_name_)->leaf_end();
    }

    //TODO
    //待完成：重叠的范围，如 id > 4, id > 5
    void update_left_right_by_cmp(CompOp cmp,bool set_lower,bool set_upper,char *lower,char *upper){
        for (const auto &cond : conds_){
            switch (cmp) {
                case OP_GT:
                    left_ = sm_manager_->ihs_.at(index_name_)->upper_bound(lower);
                    break;
                case OP_GE:
                case OP_EQ:
                    left_ = sm_manager_->ihs_.at(index_name_)->lower_bound(lower);
                    break;
                case OP_LT:
                    right_ = sm_manager_->ihs_.at(index_name_)->lower_bound(upper);
                    break;
                case OP_LE:
                    right_ = sm_manager_->ihs_.at(index_name_)->upper_bound(upper);
                    break;
            }
        }

        if (!set_lower)left_ = sm_manager_->ihs_.at(index_name_)->leaf_begin();
        if (!set_upper)right_ = sm_manager_->ihs_.at(index_name_)->leaf_end();

    }
    bool check_all_conditions(std::unique_ptr<RmRecord> &record) {
        for (const auto &cond : fed_conds_) {
            if (!check_condition(cols_, cond, record)) {
                return false;
            }
        }
        return true;
    }
    void init_current_min_max(Value &current_min,Value &current_max,const ColMeta index_col){
        int len = index_col.len;
        switch (index_col.type)
        {
            case TYPE_INT:
                current_min.set_int(INT32_MIN);
                current_max.set_int(INT32_MAX);
                break;
            case TYPE_FLOAT:
                current_min.set_float(__FLT_MIN__);
                current_max.set_float(__FLT_MAX__);
                break;
            case TYPE_STRING:
                current_min.set_str(std::string(len, 0));
                current_max.set_str(std::string(len, 127));
                break;
            default:
                break;
        }
        current_min.init_raw(len);
        current_max.init_raw(len);

    }
    bool check_condition(std::vector<ColMeta> cols, const Condition &cond, std::unique_ptr<RmRecord> &record) {
        // 没有tabs所以要传一个cols
        // 要把cond中的列的位置找出来
        auto left_col_name = cond.lhs_col.col_name;
        ColMeta left_col, right_col;
        for (const auto &col : cols) {
            if (col.name == left_col_name && col.tab_name == tab_name_) {
                left_col = col;
                break;
            }
        }
        auto left_value = get_value(record, left_col);
        // 找条件中的右边一列
        Value right_value;
        if (cond.is_rhs_val) {
            right_value = cond.rhs_val;
        } else {
            for (const auto &col : cols) {
                if (col.name == cond.rhs_col.col_name && col.tab_name == cond.rhs_col.tab_name) {
                    right_col = col;
                    break;
                }
            }
            right_value = get_value(record, right_col);
        }

        return compare_value(left_value, right_value, cond.op);
    }   

    bool compare_value(const Value& left_value,const Value& right_value,CompOp op){
        switch (op){
            case OP_EQ:
                return left_value == right_value;
            case OP_NE:
                return left_value != right_value;
            case OP_LT:
                return left_value < right_value;
            case OP_GT:
                return left_value > right_value;
            case OP_LE:
                return !(left_value > right_value);
            case OP_GE:
                return !(left_value < right_value);
        }
    }

    void beginTuple() override {
        getBound();
        auto ih = sm_manager_->ihs_.at(index_name_).get();
        scan_ = std::make_unique<IxScan>(ih, left_, right_, sm_manager_->get_bpm());
        //TODO 间隙锁添加
        //要求左开右闭？
        auto index_tree_handle = sm_manager_->ihs_.at(index_name_).get();
        //sql情况，表内id: 1 ,2 3, 4, 5,
        //case1:select id = 2, 不用上锁
        //case2:gap_lock在边界的情况:
        //case2.1:select id > 2, id < 4; left应该获取2的后一位，即第一位大于2的值,upper_bound;right应该获取第一个大于等于4的值,lower_bound
        //case2.2:select id >= 2 ,id < 4; left应该获取2,即第一位大于等于2的值,lower_bound;right，同上
        //case2.3:select id >=2 ,id <= 4;left同上，right，第一位大于4的值，5，upper_bound;
        //case2.4:一边没有数字 (如id > 2)，则为end或begin
        //上述要求应当在getbound中实现
        //注意，若通过bound得到的值的边界条件，有可能超出页面的大小，因此还需要判断

        ih->gap_lock(gap_left_,gap_right_,context_,fh_->GetFd());

        while(!scan_->is_end()) {
            // 遍历
            rid_ = scan_->rid();

            auto record = fh_->get_record(rid_, context_);

            //可能存在冗余检查，因为之前索引部分的条件依据通过过滤得到
            bool check = check_all_conditions(record);
            if (check) {
                rid_ = scan_->rid();
                return;
            }
            scan_->next();
        }
    }

    void nextTuple() override {
        scan_->next();
        while (!scan_->is_end()) {
            auto record = fh_->get_record(scan_->rid(), context_);
            bool check = check_all_conditions(record);
            if (check) {
                rid_ = scan_->rid();
                return;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (is_end()) {
            return nullptr;
        }
        //context_->lock_mgr_->lock_gap_on_index(context_->txn_,sm_manager_->ihs_.at(index_name_)->get_fd(),rid_);
        context_->lock_mgr_->lock_shared_on_record(context_->txn_,rid_,fh_->GetFd());
        return fh_->get_record(rid_, context_);
    }

    size_t tupleLen() const override { return len_; }

    std::string getType() override { return "IndexScanExecutor"; }

    const std::vector<ColMeta> &cols() const override { return cols_; }

    bool is_end() const override { return scan_->is_end(); }

    Rid &rid() override { return rid_; }
};