/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class InsertExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Value> values_;     // 需要插入的数据
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::string tab_name_;          // 表名称
    Rid rid_;                       // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
    SmManager *sm_manager_;

   public:
    InsertExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Value> values, Context *context) {
        sm_manager_ = sm_manager;
        tab_ = sm_manager_->db_.get_table(tab_name);
        values_ = values;
        tab_name_ = tab_name;
        if (values.size() != tab_.cols.size()) {
            throw InvalidValueCountError();
        }
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        context_ = context;

        if(tab_.indexes.empty()){
            context_->lock_mgr_->lock_exclusive_on_table(context_->txn_,fh_->GetFd());
        } else{
            context_->lock_mgr_->lock_IX_on_table(context_->txn_,fh_->GetFd());
        }
    };

    std::unique_ptr<RmRecord> Next() override {
        // Make record buffer
        RmRecord rec(fh_->get_file_hdr().record_size);
        for (size_t i = 0; i < values_.size(); i++) {
            auto &col = tab_.cols[i];
            auto &val = values_[i];
            if(col.type == TYPE_FLOAT && val.type == TYPE_INT){
                val.set_float(float(val.int_val));
            }else if(((col.type == TYPE_STRING && val.type != TYPE_STRING) || (col.type != TYPE_STRING && val.type == TYPE_STRING)) ) {
                throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
            }

            if (val.type == TYPE_INT && col.type == TYPE_FLOAT) {
                val.float_val = (float)val.int_val;
            } else if (val.type == TYPE_FLOAT && col.type == TYPE_INT) {
                val.int_val = (int)val.float_val;
            }
            val.type = col.type;
            val.init_raw(col.len);

            memcpy(rec.data + col.offset, val.raw->data, col.len);
        }

        // 检查索引重复
        for (size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto &index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
            char* key = new char[index.col_tot_len];
            int offset = 0;
            for(size_t j = 0; j < index.col_num; ++j) {
                memcpy(key + offset, rec.data + index.cols[j].offset, index.cols[j].len);
                offset += index.cols[j].len;
            }
            std::vector<Rid> result;
            if (ih->get_value(key, &result, context_->txn_)) {
                throw DuplicateKeyError();
            }
        }


        // Insert into record file
        rid_ = fh_->insert_record(rec.data, context_);

        //更新txn内的写集和写日志
        std::unique_ptr<InsertLogRecord> insert_log(new InsertLogRecord(context_->txn_->get_transaction_id(),context_->txn_->get_prev_lsn(),
                                                                        rec, rid_, tab_name_));
        //context_->log_mgr_->add_log_to_buffer(insert_log.get());
        context_->txn_->set_prev_lsn(insert_log->lsn_);

        auto write_record = new WriteRecord(WType::INSERT_TUPLE, tab_name_, rid_);
        context_->txn_->append_write_record(write_record);

        // Insert into index
        for(size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
            char* key = new char[index.col_tot_len];
            int offset = 0;
            for(size_t j = 0; j < index.col_num; ++j) {
                memcpy(key + offset, rec.data + index.cols[j].offset, index.cols[j].len);
                offset += index.cols[j].len;
            }
            //TODO 间隙锁
            Rid gap_rid;
            auto insert_upper_iid = ih->upper_bound(key,gap_rid);
            insert_upper_iid = {insert_upper_iid.page_no,insert_upper_iid.slot_no - 1};
            context_->lock_mgr_->try_lock_gap_on_index(context_->txn_,ih->get_fd(),gap_rid);
            ih->insert_entry(key, rid_, context_->txn_);//这里的value是rid，所以需要提前知道rid的位置信息
            context_->lock_mgr_->lock_gap_on_index(context_->txn_,ih->get_fd(),rid_);
        }

        return nullptr;
    }
    Rid &rid() override { return rid_; }
};