#include <fstream>
#include <vector>
#include <algorithm>
#include <queue>
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include <cstdio> // for std::remove

class SortMergeJoinExecutor : public AbstractExecutor {
private:
    std::unique_ptr<AbstractExecutor> left_;    // 左表
    std::unique_ptr<AbstractExecutor> right_;   // 右表
    size_t len_;                                // join后获得的每条记录的长度
    size_t left_len_;                           // 外表的记录长度
    size_t right_len_;                          // 内表的记录长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段
    std::unique_ptr<RmRecord> merged_rec;       // 合并后的记录
    std::vector<Condition> join_conds_;         // join条件
    bool is_end_;                               // 状态
    std::unique_ptr<RmRecord> left_rec_;        // 当前左表记录
    std::unique_ptr<RmRecord> right_rec_;       // 当前右表记录
    std::ifstream left_file_;
    std::ifstream right_file_;
    std::fstream outfile;
    std::string left_sorted_filename_;
    std::string right_sorted_filename_;
    bool left_end_;
    bool right_end_;
    bool is_left_index_;
    bool is_right_index_;

public:
    SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right,
                          std::vector<Condition> conds,bool is_left_index,bool is_right_index) {
        left_ = std::move(left);
        right_ = std::move(right);
        is_left_index_ = is_left_index;
        is_right_index_ = is_right_index;
        left_len_ = left_->tupleLen();
        right_len_ = right_->tupleLen();
        len_ = left_len_ + right_len_;
        cols_ = right_->cols();
        left_sorted_filename_ = "left_sorted";
        right_sorted_filename_ = "right_sorted";
        auto left_cols = left_->cols();
        for(auto &col : left_cols){
            col.offset += right_->tupleLen();
        }
        cols_.insert(cols_.end(),left_cols.begin(),left_cols.end());
        join_conds_ = std::move(conds);
        is_end_ = false;

        get_sorted_right_left_file();
        // 打开排序后的文件
        left_file_.open(left_sorted_filename_, std::ios::binary);
        right_file_.open(right_sorted_filename_, std::ios::binary);
        if (!left_file_.is_open() || !right_file_.is_open()) {
            throw std::runtime_error("Unable to open sorted files");
        }
        left_rec_ = std::make_unique<RmRecord>(left_->tupleLen());
        right_rec_ = std::make_unique<RmRecord>(right_->tupleLen());
        mergeToSortedResults();
    }

    ~SortMergeJoinExecutor(){
        if (left_file_.is_open()) left_file_.close();
        if (right_file_.is_open()) right_file_.close();
        // 删除排序后的文件
        std::remove(left_sorted_filename_.c_str());
        std::remove(right_sorted_filename_.c_str());
    }

    bool is_end() const {
        return left_end_ || right_end_;
    }

    void beginTuple() override {
        left_end_ = !left_file_.read(left_rec_->data, left_rec_->size);
        right_end_ = !right_file_.read(right_rec_->data, right_rec_->size);
        find_next_valid_rec();
    }

    void nextTuple() override {
        if (left_end_ || right_end_) return;
        left_end_ = !left_file_.read(left_rec_->data, left_rec_->size);
        find_next_valid_rec();
    }

    std::unique_ptr<RmRecord> Next() override {
        return std::make_unique<RmRecord>(this->merged_rec->size, this->merged_rec->data);
    }

    Rid &rid() override { return _abstract_rid; }

    const std::vector<ColMeta> &cols() const {
        return this->cols_;
    }

    size_t tupleLen() const {
        return this->len_;
    }

    std::string getType() { return "SortMergeJoinExecutor"; }

    ColMeta get_col_offset(const TabCol &target) override {
        for (auto iter = this->cols_.begin(); iter != this->cols_.end(); iter++) {
            if (iter->tab_name == target.tab_name && iter->name == target.col_name) {
                return *iter;
            }
        }
        return ColMeta();
    }

private:

    void get_sorted_right_left_file(){
        if (!is_left_index_) {
            sleep(1);
            externalSort(left_, left_sorted_filename_);
        } else {
            writeIndexToSortedFile(left_, left_sorted_filename_);
        }
        if (!is_right_index_) {
            sleep(1);
            externalSort(right_, right_sorted_filename_);
        } else {
            writeIndexToSortedFile(right_, right_sorted_filename_);
        }
    }
    std::unique_ptr<RmRecord> merge_record(std::unique_ptr<RmRecord>& left_record, std::unique_ptr<RmRecord>& right_record) {
        std::unique_ptr<RmRecord> ret = std::make_unique<RmRecord>(this->len_);
//        memcpy(ret->data, left_record->data, left_record->size);
//        memcpy(ret->data + left_record->size, right_record->data, right_record->size);
        memcpy(ret->data, right_record->data, right_record->size);
        memcpy(ret->data + right_record->size, left_record->data, left_record->size);
        return ret;
    }

    void find_next_valid_rec(){
        while (!left_end_ && !right_end_) {
            int cmp_result = compareRecords(left_rec_, right_rec_, join_conds_);
            if (cmp_result < 0) {
                left_end_ = !left_file_.read(left_rec_->data, left_rec_->size);
            } else if (cmp_result > 0) {
                right_end_ = !right_file_.read(right_rec_->data, right_rec_->size);
            } else {
                merged_rec = merge_record(left_rec_, right_rec_);
                break;
            }
        }
    }

    void externalSort(std::unique_ptr<AbstractExecutor>& table, const std::string& block_prefix) {
        const size_t block_size = JOINBLOCK_SIZE; // 8B
        size_t record_size = table->tupleLen();
        size_t records_per_block = block_size / record_size;

        std::vector<std::unique_ptr<RmRecord>> records;
        size_t block_count = 0;

        table->beginTuple();
        while (!table->is_end()) {
            records.push_back(table->Next());
            table->nextTuple();

            if (records.size() >= records_per_block) {
                sortAndWriteBlock(records, block_prefix + std::to_string(block_count));
                records.clear();
                block_count++;
            }
        }
        if (!records.empty()) {
            sortAndWriteBlock(records, block_prefix + std::to_string(block_count));
            block_count++;
        }

        //已经将表分成各个块了
        mergeSortedBlocks(block_prefix, block_count,record_size,block_prefix);

        //删除临时文件
        for (int i = 0;i < block_count;i++ ) {
            std::remove((block_prefix + std::to_string(i)).c_str());
        }
    }

    void sortAndWriteBlock(std::vector<std::unique_ptr<RmRecord>>& records, const std::string& filename) {
        std::sort(records.begin(), records.end(), [&](const std::unique_ptr<RmRecord>& a, const std::unique_ptr<RmRecord>& b) {
            return compareRecords(a, b, join_conds_) < 0;
        });

        std::ofstream outfile(filename, std::ios::binary);
        if (!outfile.is_open()) {
            throw std::runtime_error("Unable to open file for sorted block");
        }

        for (const auto& record : records) {
            outfile.write(record->data, record->size);
        }
        outfile.close();
    }

    void writeIndexToSortedFile(std::unique_ptr<AbstractExecutor>& table, const std::string& filename) {
        std::ofstream outfile(filename, std::ios::binary);
        if (!outfile.is_open()) {
            throw std::runtime_error("Unable to open file for sorted data");
        }

        table->beginTuple();
        while (!table->is_end()) {
            auto rec = table->Next();
            outfile.write(rec->data, rec->size);
            table->nextTuple();
        }
        outfile.close();
    }

    void mergeSortedBlocks(const std::string& block_prefix, size_t block_count, size_t record_size, const std::string& output_file) {
        std::vector<std::ifstream> block_files(block_count);
        for (size_t i = 0; i < block_count; ++i) {
            block_files[i].open(block_prefix + std::to_string(i), std::ios::binary);
            if (!block_files[i].is_open()) {
                throw std::runtime_error("Unable to open block file: " + block_prefix + std::to_string(i));
            }
        }

        std::ofstream outfile(output_file, std::ios::binary);
        if (!outfile.is_open()) {
            throw std::runtime_error("Unable to open file for sorted results");
        }

        auto cmp = [&](const std::pair<size_t, std::unique_ptr<RmRecord>>& a, const std::pair<size_t, std::unique_ptr<RmRecord>>& b) {
            return compareRecords(a.second, b.second, join_conds_) > 0;
        };
        using QueueElement = std::pair<size_t, std::unique_ptr<RmRecord>>;
        std::priority_queue<QueueElement, std::vector<QueueElement>, decltype(cmp)> pq(cmp);
        //std::priority_queue<std::pair<size_t, std::unique_ptr<RmRecord>>, std::vector<std::pair<size_t, std::unique_ptr<RmRecord>>>, decltype(cmp)> pq(cmp);

        for (size_t i = 0; i < block_count; ++i) {
            std::unique_ptr<RmRecord> rec = std::make_unique<RmRecord>(record_size);
            block_files[i].read(rec->data, rec->size);
            if (block_files[i].gcount() == record_size) {
                pq.push({i, std::move(rec)});
            }
        }
        while (!pq.empty()) {
            QueueElement elem = std::move(const_cast<QueueElement&>(pq.top()));
            pq.pop(); // 弹出顶部元素
            size_t index = elem.first;
            std::unique_ptr<RmRecord> rec = std::move(elem.second);
            outfile.write(rec->data, rec->size);

            std::unique_ptr<RmRecord> new_rec = std::make_unique<RmRecord>(record_size);
            block_files[index].read(new_rec->data, new_rec->size);
            if (block_files[index].gcount() == record_size) {
                pq.push({index, std::move(new_rec)});
            }
        }
        for (size_t i = 0; i < block_count; ++i) {
            block_files[i].close();
        }
        outfile.close();
    }

    int compareRecords(const std::unique_ptr<RmRecord>& left_rec, const std::unique_ptr<RmRecord>& right_rec, const std::vector<Condition>& conds) {
        for (const auto& cond : conds) {
            auto left_col_iter = left_->get_col(left_->cols(), cond.lhs_col);
            auto left_value = get_value(left_rec, *left_col_iter);
            Value right_value;
            if (cond.is_rhs_val) {
                right_value = cond.rhs_val;
            } else {
                auto right_col_iter = right_->get_col(right_->cols(), cond.rhs_col);
                right_value = get_value(right_rec, *right_col_iter);
            }
            int cmp_result = compare_value(left_value, right_value);
            if (cmp_result != 0) {
                return cmp_result;
            }
        }
        return 0;
    }

    int compare_value(const Value& left_value, const Value& right_value) {
        if (left_value < right_value) return -1;
        if (left_value > right_value) return 1;
        return 0;
    }

    void mergeToSortedResults(){
        outfile.open("sorted_results.txt", std::ios::out | std::ios::app);
        std::vector<std::string> columns;

        outfile << "|";
        for(auto col : right_->cols()) {
            outfile << " " << col.name << " |";
        }
        outfile << "\n";
        while(right_file_.read(right_rec_->data, right_rec_->size)){
            for (auto &col : right_->cols()) {
                std::string col_str;
                char *rec_buf = right_rec_->data + col.offset;
                if (col.type == TYPE_INT) {
                    col_str = std::to_string(*(int *)rec_buf);
                } else if (col.type == TYPE_FLOAT) {
                    col_str = std::to_string(*(float *)rec_buf);
                } else if (col.type == TYPE_STRING) {
                    col_str = std::string((char *)rec_buf, col.len);
                    col_str.resize(strlen(col_str.c_str()));
                }
                columns.push_back(col_str);
            }
            outfile << "|";
            for(int i = 0; i < columns.size(); ++i) {
                outfile << " " << columns[i] << " |";
            }
            columns.clear();
            outfile << "\n";
        }
        right_file_.clear();
        right_file_.seekg(0,std::ios::beg);

        outfile << "|";
        for(auto col : left_->cols()) {
            outfile << " " << col.name << " |";
        }
        outfile << "\n";
        while(left_file_.read(left_rec_->data, left_rec_->size)){
            for (auto &col : left_->cols()) {
                std::string col_str;
                char *rec_buf = left_rec_->data + col.offset;
                if (col.type == TYPE_INT) {
                    col_str = std::to_string(*(int *)rec_buf);
                } else if (col.type == TYPE_FLOAT) {
                    col_str = std::to_string(*(float *)rec_buf);
                } else if (col.type == TYPE_STRING) {
                    col_str = std::string((char *)rec_buf, col.len);
                    col_str.resize(strlen(col_str.c_str()));
                }
                columns.push_back(col_str);
            }
            outfile << "|";
            for(int i = 0; i < columns.size(); ++i) {
                outfile << " " << columns[i] << " |";
            }
            columns.clear();
            outfile << "\n";
        }
        left_file_.clear();
        left_file_.seekg(0,std::ios::beg);
        outfile.close();
    }
};

