/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include <cstring>
#include <thread>
#include "log_manager.h"

/**
 * @description: 添加日志记录到日志缓冲区中，并返回日志记录号
 * @param {LogRecord*} log_record 要写入缓冲区的日志记录
 * @return {lsn_t} 返回该日志的日志记录号
 */

lsn_t LogManager::add_log_to_buffer(LogRecord* log_record) {
    //先判断是否满，如果满了，则阻塞，等待buffer刷到磁盘后再写，这里用cv，用offset
    //若没有满，则给出对应的日志号，并维护相应信息
    latch_.lock();
    auto r_size = log_record->log_tot_len_;
    if(log_buffer_->is_full(r_size)){
        latch_.unlock();
//        need_flush_ = true;
//        //需要线程刷盘,需要保证它已经刷盘完成在等待
//        //is_flushing_cv_.wait(l,[&]{return !is_flushing_;});
//        need_flush_cv_.notify_one();
//        while(log_buffer_->is_full(r_size)){
//            swap_cv_.wait(l,[&]{return !log_buffer_->is_full(r_size);});
//        }
        flush_log_to_disk();
        latch_.lock();
    }

    log_record->lsn_ = global_lsn_++;//
    //转为字节形式
    log_record->serialize(log_buffer_->buffer_ + log_buffer_->offset_);
    log_buffer_->offset_ += log_record->log_tot_len_;
    latch_.unlock();

    return log_record->lsn_;
}

/**
 * @description: 把日志缓冲区的内容刷到磁盘中，由于目前只设置了一个缓冲区，因此需要阻塞其他日志操作
 */
void LogManager::flush_log_to_disk() {
    std::unique_lock<std::mutex>l(latch_);
    //没有日志数据时直接返回
    if(log_buffer_->offset_ == 0){
        return;
    }
    persist_lsn_ = global_lsn_ - 1;
    disk_manager_->write_log(log_buffer_->buffer_,log_buffer_->offset_);
    log_buffer_->offset_ = 0;
}

/**
 * @description:将日志缓冲区交换到flush_buffer
 */
void LogManager::flushOnce() {
    if(log_buffer_->offset_ == 0){
        return;
    }
    // 互换 log buffer 和 flush buffer
    std::swap(log_buffer_, flush_buffer_);
    std::swap(log_buffer_->offset_, flush_buffer_->offset_);
    // 将 flush buffer 写入磁盘，并重置 flush buffer offset
    disk_manager_->write_log(flush_buffer_->buffer_,flush_buffer_->offset_);
    flush_buffer_->offset_ = 0;
    // 更新 persisten_lsn，表示已写入磁盘的日志的序列号
    persist_lsn_ = global_lsn_ - 1;

}

void LogManager::RunFlushThread() {
    if(!enable_logging){
        return;
    }
    flush_thread_ = new std::thread([&]{
        //防止 flush 线程和追加日志线程同时运行导致冲突
        while(true){
            std::unique_lock<std::mutex> l(latch_);
            //每隔LOG_TIMEOUT刷新一次或者buffer已满或者强制刷盘
            //is_flushing_ = false;
            //is_flushing_cv_.notify_all();
            need_flush_cv_.wait_for(l, LOG_FLUSH_TIME, [&] {return need_flush_.load();});
            if(!enable_logging)return ;
            if(log_buffer_->offset_ == 0)continue;
            std::swap(log_buffer_, flush_buffer_);
            std::swap(log_buffer_->offset_, flush_buffer_->offset_);
            l.unlock();//追加日志和flush线程可以并行运行
            //this->flushOnce();
            swap_cv_.notify_all();
            //is_flushing_ = true;
            flush_log_to_disk();
            //flushed_once_cv_.notify_all();//完成flush
            need_flush_ = false;
        }
    });
}

void LogManager::flush_dirty_page(lsn_t lsn) {
    if(persist_lsn_ >= lsn)return;
    flush_log_to_disk();
    assert(persist_lsn_ >= lsn);
    //while(lsn > persist_lsn_){};//TODO cv
}
